
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#pragma once

#include <glib-object.h>

G_BEGIN_DECLS
/* enumerations from "grd-enums.h" */
GType grd_rdp_screen_share_mode_get_type (void) G_GNUC_CONST;
#define GRD_TYPE_RDP_SCREEN_SHARE_MODE (grd_rdp_screen_share_mode_get_type ())
GType grd_vnc_auth_method_get_type (void) G_GNUC_CONST;
#define GRD_TYPE_VNC_AUTH_METHOD (grd_vnc_auth_method_get_type ())
GType grd_vnc_screen_share_mode_get_type (void) G_GNUC_CONST;
#define GRD_TYPE_VNC_SCREEN_SHARE_MODE (grd_vnc_screen_share_mode_get_type ())
GType grd_runtime_mode_get_type (void) G_GNUC_CONST;
#define GRD_TYPE_RUNTIME_MODE (grd_runtime_mode_get_type ())
GType grd_rdp_auth_methods_get_type (void) G_GNUC_CONST;
#define GRD_TYPE_RDP_AUTH_METHODS (grd_rdp_auth_methods_get_type ())
G_END_DECLS

/* Generated data ends here */

