
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */


#include <glib-object.h>

/* enumerations from "/build/gnome-remote-desktop/src/build/../gnome-remote-desktop/src/grd-enums.h" */
#include "/build/gnome-remote-desktop/src/build/../gnome-remote-desktop/src/grd-enums.h"
GType
grd_rdp_screen_share_mode_get_type (void)
{
  static size_t g_enum_type_id = 0;

  if (g_once_init_enter (&g_enum_type_id))
    {
      static const GEnumValue values[] = {
        { GRD_RDP_SCREEN_SHARE_MODE_MIRROR_PRIMARY, "GRD_RDP_SCREEN_SHARE_MODE_MIRROR_PRIMARY", "mirror-primary" },
        { GRD_RDP_SCREEN_SHARE_MODE_EXTEND, "GRD_RDP_SCREEN_SHARE_MODE_EXTEND", "extend" },
        { 0, NULL, NULL }
      };
      GType id;

      id = g_enum_register_static (g_intern_static_string ("GrdRdpScreenShareMode"), values);

      g_once_init_leave (&g_enum_type_id, id);
    }

  return g_enum_type_id;
}
GType
grd_vnc_auth_method_get_type (void)
{
  static size_t g_enum_type_id = 0;

  if (g_once_init_enter (&g_enum_type_id))
    {
      static const GEnumValue values[] = {
        { GRD_VNC_AUTH_METHOD_PROMPT, "GRD_VNC_AUTH_METHOD_PROMPT", "prompt" },
        { GRD_VNC_AUTH_METHOD_PASSWORD, "GRD_VNC_AUTH_METHOD_PASSWORD", "password" },
        { 0, NULL, NULL }
      };
      GType id;

      id = g_enum_register_static (g_intern_static_string ("GrdVncAuthMethod"), values);

      g_once_init_leave (&g_enum_type_id, id);
    }

  return g_enum_type_id;
}
GType
grd_vnc_screen_share_mode_get_type (void)
{
  static size_t g_enum_type_id = 0;

  if (g_once_init_enter (&g_enum_type_id))
    {
      static const GEnumValue values[] = {
        { GRD_VNC_SCREEN_SHARE_MODE_MIRROR_PRIMARY, "GRD_VNC_SCREEN_SHARE_MODE_MIRROR_PRIMARY", "mirror-primary" },
        { GRD_VNC_SCREEN_SHARE_MODE_EXTEND, "GRD_VNC_SCREEN_SHARE_MODE_EXTEND", "extend" },
        { 0, NULL, NULL }
      };
      GType id;

      id = g_enum_register_static (g_intern_static_string ("GrdVncScreenShareMode"), values);

      g_once_init_leave (&g_enum_type_id, id);
    }

  return g_enum_type_id;
}
GType
grd_runtime_mode_get_type (void)
{
  static size_t g_enum_type_id = 0;

  if (g_once_init_enter (&g_enum_type_id))
    {
      static const GEnumValue values[] = {
        { GRD_RUNTIME_MODE_SCREEN_SHARE, "GRD_RUNTIME_MODE_SCREEN_SHARE", "screen-share" },
        { GRD_RUNTIME_MODE_HEADLESS, "GRD_RUNTIME_MODE_HEADLESS", "headless" },
        { GRD_RUNTIME_MODE_SYSTEM, "GRD_RUNTIME_MODE_SYSTEM", "system" },
        { GRD_RUNTIME_MODE_HANDOVER, "GRD_RUNTIME_MODE_HANDOVER", "handover" },
        { 0, NULL, NULL }
      };
      GType id;

      id = g_enum_register_static (g_intern_static_string ("GrdRuntimeMode"), values);

      g_once_init_leave (&g_enum_type_id, id);
    }

  return g_enum_type_id;
}
GType
grd_rdp_auth_methods_get_type (void)
{
  static size_t g_enum_type_id = 0;

  if (g_once_init_enter (&g_enum_type_id))
    {
      static const GFlagsValue values[] = {
        { GRD_RDP_AUTH_METHOD_CREDENTIALS, "GRD_RDP_AUTH_METHOD_CREDENTIALS", "credentials" },
        { GRD_RDP_AUTH_METHOD_KERBEROS, "GRD_RDP_AUTH_METHOD_KERBEROS", "kerberos" },
        { 0, NULL, NULL }
      };
      GType id;

      id = g_flags_register_static (g_intern_static_string ("GrdRdpAuthMethods"), values);

      g_once_init_leave (&g_enum_type_id, id);
    }

  return g_enum_type_id;
}

/* Generated data ends here */

