/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __GRD_DBUS_GDM_H__
#define __GRD_DBUS_GDM_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.DisplayManager.RemoteDisplay */

#define GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY (grd_dbus_gdm_remote_display_get_type ())
#define GRD_DBUS_GDM_REMOTE_DISPLAY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY, GrdDBusGdmRemoteDisplay))
#define GRD_DBUS_GDM_IS_REMOTE_DISPLAY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY))
#define GRD_DBUS_GDM_REMOTE_DISPLAY_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY, GrdDBusGdmRemoteDisplayIface))

struct _GrdDBusGdmRemoteDisplay;
typedef struct _GrdDBusGdmRemoteDisplay GrdDBusGdmRemoteDisplay;
typedef struct _GrdDBusGdmRemoteDisplayIface GrdDBusGdmRemoteDisplayIface;

struct _GrdDBusGdmRemoteDisplayIface
{
  GTypeInterface parent_iface;

  const gchar * (*get_remote_id) (GrdDBusGdmRemoteDisplay *object);

  const gchar * (*get_session_id) (GrdDBusGdmRemoteDisplay *object);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdDBusGdmRemoteDisplay, g_object_unref)
#endif

GType grd_dbus_gdm_remote_display_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *grd_dbus_gdm_remote_display_interface_info (void);
guint grd_dbus_gdm_remote_display_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus property accessors: */
const gchar *grd_dbus_gdm_remote_display_get_remote_id (GrdDBusGdmRemoteDisplay *object);
gchar *grd_dbus_gdm_remote_display_dup_remote_id (GrdDBusGdmRemoteDisplay *object);
void grd_dbus_gdm_remote_display_set_remote_id (GrdDBusGdmRemoteDisplay *object, const gchar *value);

const gchar *grd_dbus_gdm_remote_display_get_session_id (GrdDBusGdmRemoteDisplay *object);
gchar *grd_dbus_gdm_remote_display_dup_session_id (GrdDBusGdmRemoteDisplay *object);
void grd_dbus_gdm_remote_display_set_session_id (GrdDBusGdmRemoteDisplay *object, const gchar *value);


/* ---- */

#define GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_PROXY (grd_dbus_gdm_remote_display_proxy_get_type ())
#define GRD_DBUS_GDM_REMOTE_DISPLAY_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_PROXY, GrdDBusGdmRemoteDisplayProxy))
#define GRD_DBUS_GDM_REMOTE_DISPLAY_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_PROXY, GrdDBusGdmRemoteDisplayProxyClass))
#define GRD_DBUS_GDM_REMOTE_DISPLAY_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_PROXY, GrdDBusGdmRemoteDisplayProxyClass))
#define GRD_DBUS_GDM_IS_REMOTE_DISPLAY_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_PROXY))
#define GRD_DBUS_GDM_IS_REMOTE_DISPLAY_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_PROXY))

typedef struct _GrdDBusGdmRemoteDisplayProxy GrdDBusGdmRemoteDisplayProxy;
typedef struct _GrdDBusGdmRemoteDisplayProxyClass GrdDBusGdmRemoteDisplayProxyClass;
typedef struct _GrdDBusGdmRemoteDisplayProxyPrivate GrdDBusGdmRemoteDisplayProxyPrivate;

struct _GrdDBusGdmRemoteDisplayProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  GrdDBusGdmRemoteDisplayProxyPrivate *priv;
};

struct _GrdDBusGdmRemoteDisplayProxyClass
{
  GDBusProxyClass parent_class;
};

GType grd_dbus_gdm_remote_display_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdDBusGdmRemoteDisplayProxy, g_object_unref)
#endif

void grd_dbus_gdm_remote_display_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GrdDBusGdmRemoteDisplay *grd_dbus_gdm_remote_display_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
GrdDBusGdmRemoteDisplay *grd_dbus_gdm_remote_display_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void grd_dbus_gdm_remote_display_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GrdDBusGdmRemoteDisplay *grd_dbus_gdm_remote_display_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
GrdDBusGdmRemoteDisplay *grd_dbus_gdm_remote_display_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_SKELETON (grd_dbus_gdm_remote_display_skeleton_get_type ())
#define GRD_DBUS_GDM_REMOTE_DISPLAY_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_SKELETON, GrdDBusGdmRemoteDisplaySkeleton))
#define GRD_DBUS_GDM_REMOTE_DISPLAY_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_SKELETON, GrdDBusGdmRemoteDisplaySkeletonClass))
#define GRD_DBUS_GDM_REMOTE_DISPLAY_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_SKELETON, GrdDBusGdmRemoteDisplaySkeletonClass))
#define GRD_DBUS_GDM_IS_REMOTE_DISPLAY_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_SKELETON))
#define GRD_DBUS_GDM_IS_REMOTE_DISPLAY_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_SKELETON))

typedef struct _GrdDBusGdmRemoteDisplaySkeleton GrdDBusGdmRemoteDisplaySkeleton;
typedef struct _GrdDBusGdmRemoteDisplaySkeletonClass GrdDBusGdmRemoteDisplaySkeletonClass;
typedef struct _GrdDBusGdmRemoteDisplaySkeletonPrivate GrdDBusGdmRemoteDisplaySkeletonPrivate;

struct _GrdDBusGdmRemoteDisplaySkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  GrdDBusGdmRemoteDisplaySkeletonPrivate *priv;
};

struct _GrdDBusGdmRemoteDisplaySkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType grd_dbus_gdm_remote_display_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdDBusGdmRemoteDisplaySkeleton, g_object_unref)
#endif

GrdDBusGdmRemoteDisplay *grd_dbus_gdm_remote_display_skeleton_new (void);


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.DisplayManager.RemoteDisplayFactory */

#define GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY (grd_dbus_gdm_remote_display_factory_get_type ())
#define GRD_DBUS_GDM_REMOTE_DISPLAY_FACTORY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY, GrdDBusGdmRemoteDisplayFactory))
#define GRD_DBUS_GDM_IS_REMOTE_DISPLAY_FACTORY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY))
#define GRD_DBUS_GDM_REMOTE_DISPLAY_FACTORY_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY, GrdDBusGdmRemoteDisplayFactoryIface))

struct _GrdDBusGdmRemoteDisplayFactory;
typedef struct _GrdDBusGdmRemoteDisplayFactory GrdDBusGdmRemoteDisplayFactory;
typedef struct _GrdDBusGdmRemoteDisplayFactoryIface GrdDBusGdmRemoteDisplayFactoryIface;

struct _GrdDBusGdmRemoteDisplayFactoryIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_create_remote_display) (
    GrdDBusGdmRemoteDisplayFactory *object,
    GDBusMethodInvocation *invocation,
    GVariant *arg_properties);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdDBusGdmRemoteDisplayFactory, g_object_unref)
#endif

GType grd_dbus_gdm_remote_display_factory_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *grd_dbus_gdm_remote_display_factory_interface_info (void);
guint grd_dbus_gdm_remote_display_factory_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void grd_dbus_gdm_remote_display_factory_complete_create_remote_display (
    GrdDBusGdmRemoteDisplayFactory *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void grd_dbus_gdm_remote_display_factory_call_create_remote_display (
    GrdDBusGdmRemoteDisplayFactory *proxy,
    GVariant *arg_properties,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean grd_dbus_gdm_remote_display_factory_call_create_remote_display_finish (
    GrdDBusGdmRemoteDisplayFactory *proxy,
    GAsyncResult *res,
    GError **error);

gboolean grd_dbus_gdm_remote_display_factory_call_create_remote_display_sync (
    GrdDBusGdmRemoteDisplayFactory *proxy,
    GVariant *arg_properties,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY_PROXY (grd_dbus_gdm_remote_display_factory_proxy_get_type ())
#define GRD_DBUS_GDM_REMOTE_DISPLAY_FACTORY_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY_PROXY, GrdDBusGdmRemoteDisplayFactoryProxy))
#define GRD_DBUS_GDM_REMOTE_DISPLAY_FACTORY_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY_PROXY, GrdDBusGdmRemoteDisplayFactoryProxyClass))
#define GRD_DBUS_GDM_REMOTE_DISPLAY_FACTORY_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY_PROXY, GrdDBusGdmRemoteDisplayFactoryProxyClass))
#define GRD_DBUS_GDM_IS_REMOTE_DISPLAY_FACTORY_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY_PROXY))
#define GRD_DBUS_GDM_IS_REMOTE_DISPLAY_FACTORY_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY_PROXY))

typedef struct _GrdDBusGdmRemoteDisplayFactoryProxy GrdDBusGdmRemoteDisplayFactoryProxy;
typedef struct _GrdDBusGdmRemoteDisplayFactoryProxyClass GrdDBusGdmRemoteDisplayFactoryProxyClass;
typedef struct _GrdDBusGdmRemoteDisplayFactoryProxyPrivate GrdDBusGdmRemoteDisplayFactoryProxyPrivate;

struct _GrdDBusGdmRemoteDisplayFactoryProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  GrdDBusGdmRemoteDisplayFactoryProxyPrivate *priv;
};

struct _GrdDBusGdmRemoteDisplayFactoryProxyClass
{
  GDBusProxyClass parent_class;
};

GType grd_dbus_gdm_remote_display_factory_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdDBusGdmRemoteDisplayFactoryProxy, g_object_unref)
#endif

void grd_dbus_gdm_remote_display_factory_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GrdDBusGdmRemoteDisplayFactory *grd_dbus_gdm_remote_display_factory_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
GrdDBusGdmRemoteDisplayFactory *grd_dbus_gdm_remote_display_factory_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void grd_dbus_gdm_remote_display_factory_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GrdDBusGdmRemoteDisplayFactory *grd_dbus_gdm_remote_display_factory_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
GrdDBusGdmRemoteDisplayFactory *grd_dbus_gdm_remote_display_factory_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY_SKELETON (grd_dbus_gdm_remote_display_factory_skeleton_get_type ())
#define GRD_DBUS_GDM_REMOTE_DISPLAY_FACTORY_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY_SKELETON, GrdDBusGdmRemoteDisplayFactorySkeleton))
#define GRD_DBUS_GDM_REMOTE_DISPLAY_FACTORY_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY_SKELETON, GrdDBusGdmRemoteDisplayFactorySkeletonClass))
#define GRD_DBUS_GDM_REMOTE_DISPLAY_FACTORY_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY_SKELETON, GrdDBusGdmRemoteDisplayFactorySkeletonClass))
#define GRD_DBUS_GDM_IS_REMOTE_DISPLAY_FACTORY_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY_SKELETON))
#define GRD_DBUS_GDM_IS_REMOTE_DISPLAY_FACTORY_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GRD_DBUS_GDM_TYPE_REMOTE_DISPLAY_FACTORY_SKELETON))

typedef struct _GrdDBusGdmRemoteDisplayFactorySkeleton GrdDBusGdmRemoteDisplayFactorySkeleton;
typedef struct _GrdDBusGdmRemoteDisplayFactorySkeletonClass GrdDBusGdmRemoteDisplayFactorySkeletonClass;
typedef struct _GrdDBusGdmRemoteDisplayFactorySkeletonPrivate GrdDBusGdmRemoteDisplayFactorySkeletonPrivate;

struct _GrdDBusGdmRemoteDisplayFactorySkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  GrdDBusGdmRemoteDisplayFactorySkeletonPrivate *priv;
};

struct _GrdDBusGdmRemoteDisplayFactorySkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType grd_dbus_gdm_remote_display_factory_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdDBusGdmRemoteDisplayFactorySkeleton, g_object_unref)
#endif

GrdDBusGdmRemoteDisplayFactory *grd_dbus_gdm_remote_display_factory_skeleton_new (void);


/* ---- */

#define GRD_DBUS_GDM_TYPE_OBJECT (grd_dbus_gdm_object_get_type ())
#define GRD_DBUS_GDM_OBJECT(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GRD_DBUS_GDM_TYPE_OBJECT, GrdDBusGdmObject))
#define GRD_DBUS_GDM_IS_OBJECT(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GRD_DBUS_GDM_TYPE_OBJECT))
#define GRD_DBUS_GDM_OBJECT_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), GRD_DBUS_GDM_TYPE_OBJECT, GrdDBusGdmObject))

struct _GrdDBusGdmObject;
typedef struct _GrdDBusGdmObject GrdDBusGdmObject;
typedef struct _GrdDBusGdmObjectIface GrdDBusGdmObjectIface;

struct _GrdDBusGdmObjectIface
{
  GTypeInterface parent_iface;
};

GType grd_dbus_gdm_object_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdDBusGdmObject, g_object_unref)
#endif

GrdDBusGdmRemoteDisplay *grd_dbus_gdm_object_get_remote_display (GrdDBusGdmObject *object);
GrdDBusGdmRemoteDisplayFactory *grd_dbus_gdm_object_get_remote_display_factory (GrdDBusGdmObject *object);
GrdDBusGdmRemoteDisplay *grd_dbus_gdm_object_peek_remote_display (GrdDBusGdmObject *object);
GrdDBusGdmRemoteDisplayFactory *grd_dbus_gdm_object_peek_remote_display_factory (GrdDBusGdmObject *object);

#define GRD_DBUS_GDM_TYPE_OBJECT_PROXY (grd_dbus_gdm_object_proxy_get_type ())
#define GRD_DBUS_GDM_OBJECT_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GRD_DBUS_GDM_TYPE_OBJECT_PROXY, GrdDBusGdmObjectProxy))
#define GRD_DBUS_GDM_OBJECT_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GRD_DBUS_GDM_TYPE_OBJECT_PROXY, GrdDBusGdmObjectProxyClass))
#define GRD_DBUS_GDM_OBJECT_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GRD_DBUS_GDM_TYPE_OBJECT_PROXY, GrdDBusGdmObjectProxyClass))
#define GRD_DBUS_GDM_IS_OBJECT_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GRD_DBUS_GDM_TYPE_OBJECT_PROXY))
#define GRD_DBUS_GDM_IS_OBJECT_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GRD_DBUS_GDM_TYPE_OBJECT_PROXY))

typedef struct _GrdDBusGdmObjectProxy GrdDBusGdmObjectProxy;
typedef struct _GrdDBusGdmObjectProxyClass GrdDBusGdmObjectProxyClass;
typedef struct _GrdDBusGdmObjectProxyPrivate GrdDBusGdmObjectProxyPrivate;

struct _GrdDBusGdmObjectProxy
{
  /*< private >*/
  GDBusObjectProxy parent_instance;
  GrdDBusGdmObjectProxyPrivate *priv;
};

struct _GrdDBusGdmObjectProxyClass
{
  GDBusObjectProxyClass parent_class;
};

GType grd_dbus_gdm_object_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdDBusGdmObjectProxy, g_object_unref)
#endif

GrdDBusGdmObjectProxy *grd_dbus_gdm_object_proxy_new (GDBusConnection *connection, const gchar *object_path);

#define GRD_DBUS_GDM_TYPE_OBJECT_SKELETON (grd_dbus_gdm_object_skeleton_get_type ())
#define GRD_DBUS_GDM_OBJECT_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GRD_DBUS_GDM_TYPE_OBJECT_SKELETON, GrdDBusGdmObjectSkeleton))
#define GRD_DBUS_GDM_OBJECT_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GRD_DBUS_GDM_TYPE_OBJECT_SKELETON, GrdDBusGdmObjectSkeletonClass))
#define GRD_DBUS_GDM_OBJECT_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GRD_DBUS_GDM_TYPE_OBJECT_SKELETON, GrdDBusGdmObjectSkeletonClass))
#define GRD_DBUS_GDM_IS_OBJECT_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GRD_DBUS_GDM_TYPE_OBJECT_SKELETON))
#define GRD_DBUS_GDM_IS_OBJECT_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GRD_DBUS_GDM_TYPE_OBJECT_SKELETON))

typedef struct _GrdDBusGdmObjectSkeleton GrdDBusGdmObjectSkeleton;
typedef struct _GrdDBusGdmObjectSkeletonClass GrdDBusGdmObjectSkeletonClass;
typedef struct _GrdDBusGdmObjectSkeletonPrivate GrdDBusGdmObjectSkeletonPrivate;

struct _GrdDBusGdmObjectSkeleton
{
  /*< private >*/
  GDBusObjectSkeleton parent_instance;
  GrdDBusGdmObjectSkeletonPrivate *priv;
};

struct _GrdDBusGdmObjectSkeletonClass
{
  GDBusObjectSkeletonClass parent_class;
};

GType grd_dbus_gdm_object_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdDBusGdmObjectSkeleton, g_object_unref)
#endif

GrdDBusGdmObjectSkeleton *grd_dbus_gdm_object_skeleton_new (const gchar *object_path);
void grd_dbus_gdm_object_skeleton_set_remote_display (GrdDBusGdmObjectSkeleton *object, GrdDBusGdmRemoteDisplay *interface_);
void grd_dbus_gdm_object_skeleton_set_remote_display_factory (GrdDBusGdmObjectSkeleton *object, GrdDBusGdmRemoteDisplayFactory *interface_);

/* ---- */

#define GRD_DBUS_GDM_TYPE_OBJECT_MANAGER_CLIENT (grd_dbus_gdm_object_manager_client_get_type ())
#define GRD_DBUS_GDM_OBJECT_MANAGER_CLIENT(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GRD_DBUS_GDM_TYPE_OBJECT_MANAGER_CLIENT, GrdDBusGdmObjectManagerClient))
#define GRD_DBUS_GDM_OBJECT_MANAGER_CLIENT_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GRD_DBUS_GDM_TYPE_OBJECT_MANAGER_CLIENT, GrdDBusGdmObjectManagerClientClass))
#define GRD_DBUS_GDM_OBJECT_MANAGER_CLIENT_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GRD_DBUS_GDM_TYPE_OBJECT_MANAGER_CLIENT, GrdDBusGdmObjectManagerClientClass))
#define GRD_DBUS_GDM_IS_OBJECT_MANAGER_CLIENT(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GRD_DBUS_GDM_TYPE_OBJECT_MANAGER_CLIENT))
#define GRD_DBUS_GDM_IS_OBJECT_MANAGER_CLIENT_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GRD_DBUS_GDM_TYPE_OBJECT_MANAGER_CLIENT))

typedef struct _GrdDBusGdmObjectManagerClient GrdDBusGdmObjectManagerClient;
typedef struct _GrdDBusGdmObjectManagerClientClass GrdDBusGdmObjectManagerClientClass;
typedef struct _GrdDBusGdmObjectManagerClientPrivate GrdDBusGdmObjectManagerClientPrivate;

struct _GrdDBusGdmObjectManagerClient
{
  /*< private >*/
  GDBusObjectManagerClient parent_instance;
  GrdDBusGdmObjectManagerClientPrivate *priv;
};

struct _GrdDBusGdmObjectManagerClientClass
{
  GDBusObjectManagerClientClass parent_class;
};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdDBusGdmObjectManagerClient, g_object_unref)
#endif

GType grd_dbus_gdm_object_manager_client_get_type (void) G_GNUC_CONST;

GType grd_dbus_gdm_object_manager_client_get_proxy_type (GDBusObjectManagerClient *manager, const gchar *object_path, const gchar *interface_name, gpointer user_data);

void grd_dbus_gdm_object_manager_client_new (
    GDBusConnection        *connection,
    GDBusObjectManagerClientFlags  flags,
    const gchar            *name,
    const gchar            *object_path,
    GCancellable           *cancellable,
    GAsyncReadyCallback     callback,
    gpointer                user_data);
GDBusObjectManager *grd_dbus_gdm_object_manager_client_new_finish (
    GAsyncResult        *res,
    GError             **error);
GDBusObjectManager *grd_dbus_gdm_object_manager_client_new_sync (
    GDBusConnection        *connection,
    GDBusObjectManagerClientFlags  flags,
    const gchar            *name,
    const gchar            *object_path,
    GCancellable           *cancellable,
    GError                **error);

void grd_dbus_gdm_object_manager_client_new_for_bus (
    GBusType                bus_type,
    GDBusObjectManagerClientFlags  flags,
    const gchar            *name,
    const gchar            *object_path,
    GCancellable           *cancellable,
    GAsyncReadyCallback     callback,
    gpointer                user_data);
GDBusObjectManager *grd_dbus_gdm_object_manager_client_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
GDBusObjectManager *grd_dbus_gdm_object_manager_client_new_for_bus_sync (
    GBusType                bus_type,
    GDBusObjectManagerClientFlags  flags,
    const gchar            *name,
    const gchar            *object_path,
    GCancellable           *cancellable,
    GError                **error);


G_END_DECLS

#endif /* __GRD_DBUS_GDM_H__ */
